import { FederatedPointerEvent } from 'pixi.js';

/**
 * Button states for visual feedback
 */
export declare enum ButtonState {
    Normal = "normal",
    Hover = "hover",
    Pressed = "pressed",
    Disabled = "disabled"
}
/**
 * Button style configuration for different visual approaches
 */
export interface ButtonStyle {
    /** Background color for each state */
    backgroundColor?: {
        [ButtonState.Normal]?: string;
        [ButtonState.Hover]?: string;
        [ButtonState.Pressed]?: string;
        [ButtonState.Disabled]?: string;
    };
    /** Border configuration */
    border?: {
        color?: string;
        width?: number;
        radius?: number;
    };
    /** Text styling */
    text?: {
        color?: string;
        fontSize?: number;
        fontFamily?: string;
    };
    /** Sprite textures for each state (alternative to backgroundColor) */
    textures?: {
        [ButtonState.Normal]?: string;
        [ButtonState.Hover]?: string;
        [ButtonState.Pressed]?: string;
        [ButtonState.Disabled]?: string;
    };
}
/**
 * Properties for the Button component
 */
export interface ButtonProps {
    /** Button text content */
    text?: string;
    /** Button disabled state */
    disabled?: boolean;
    /** Click event handler */
    click?: (event: FederatedPointerEvent) => void;
    /** Hover enter event handler */
    hoverEnter?: (event: FederatedPointerEvent) => void;
    /** Hover leave event handler */
    hoverLeave?: (event: FederatedPointerEvent) => void;
    /** Press down event handler */
    pressDown?: (event: FederatedPointerEvent) => void;
    /** Press up event handler */
    pressUp?: (event: FederatedPointerEvent) => void;
    /** Visual style configuration */
    style?: ButtonStyle;
    /** Button width */
    width?: number;
    /** Button height */
    height?: number;
    /** Button position X */
    x?: number;
    /** Button position Y */
    y?: number;
    /** Button alpha/opacity */
    alpha?: number;
    /** Button visibility */
    visible?: boolean;
    /** Button cursor */
    cursor?: string;
}
/**
 * Creates a Button component with interactive states and customizable styling.
 *
 * This component provides a fully interactive button with visual feedback
 * for different states (normal, hover, pressed, disabled). It supports both
 * sprite-based and graphics-based rendering approaches.
 *
 * The button is built using a Container with background and text elements,
 * providing reactive state management and event handling.
 *
 * @param props - Button configuration including text, styling, and event handlers
 * @returns A reactive Button component
 * @example
 * ```typescript
 * // Simple button with text and click handler
 * const simpleButton = Button({
 *   text: "Click Me",
 *   onClick: () => console.log("Button clicked!"),
 *   width: 150,
 *   height: 50
 * });
 *
 * // Styled button with custom colors
 * const styledButton = Button({
 *   text: "Styled Button",
 *   style: {
 *     backgroundColor: {
 *       normal: "#28a745",
 *       hover: "#218838",
 *       pressed: "#1e7e34",
 *       disabled: "#6c757d"
 *     },
 *     border: {
 *       radius: 8,
 *       width: 2,
 *       color: "#ffffff"
 *     },
 *     text: {
 *       fontSize: 18,
 *       color: "#ffffff"
 *     }
 *   }
 * });
 *
 * // Sprite-based button
 * const spriteButton = Button({
 *   text: "Play Game",
 *   style: {
 *     textures: {
 *       normal: "/assets/button-normal.png",
 *       hover: "/assets/button-hover.png",
 *       pressed: "/assets/button-pressed.png"
 *     }
 *   }
 * });
 * ```
 */
export declare function Button(props: ButtonProps): import('..').Element<import('./DisplayObject').ComponentInstance> | Promise<import('..').Element<import('./DisplayObject').ComponentInstance>>;
//# sourceMappingURL=Button.d.ts.map