import { Graphics as PixiGraphics } from 'pixi.js';
import { Element } from '../engine/reactive';
import { ComponentInstance } from './DisplayObject';
import { DisplayObjectProps } from './types/DisplayObject';
import { SignalOrPrimitive } from './types';

interface GraphicsProps extends DisplayObjectProps {
    draw?: (graphics: PixiGraphics, width: number, height: number) => void;
}
interface RectProps extends DisplayObjectProps {
    color: SignalOrPrimitive<string>;
}
interface CircleProps extends DisplayObjectProps {
    radius: SignalOrPrimitive<number>;
    color: SignalOrPrimitive<string>;
}
interface EllipseProps extends DisplayObjectProps {
    color: SignalOrPrimitive<string>;
}
interface TriangleProps extends DisplayObjectProps {
    base: SignalOrPrimitive<number>;
    color: SignalOrPrimitive<string>;
}
interface SvgProps extends DisplayObjectProps {
    /** SVG content as string (legacy prop) */
    svg?: string;
    /** URL source of the SVG file to load */
    src?: string;
    /** Direct SVG content as string */
    content?: string;
}
export declare function Graphics(props: GraphicsProps): Element<ComponentInstance>;
export declare function Rect(props: RectProps): Element<ComponentInstance>;
export declare function Circle(props: CircleProps): Element<ComponentInstance>;
export declare function Ellipse(props: EllipseProps): Element<ComponentInstance>;
export declare function Triangle(props: TriangleProps): Element<ComponentInstance>;
/**
 * Creates an SVG component that can render SVG graphics from URL, content, or legacy svg prop.
 *
 * This component provides three ways to display SVG graphics:
 * - **src**: Load SVG from a URL using Assets.load with parseAsGraphicsContext option
 * - **content**: Render SVG directly from string content using Graphics.svg() method
 * - **svg**: Legacy prop for SVG content (for backward compatibility)
 *
 * @param props - Component properties including src, content, or svg
 * @returns A reactive SVG component
 * @example
 * ```typescript
 * // Load from URL
 * const svgFromUrl = Svg({ src: "/assets/logo.svg" });
 *
 * // Direct content
 * const svgFromContent = Svg({
 *   content: `<svg viewBox="0 0 100 100">
 *     <circle cx="50" cy="50" r="40" fill="blue"/>
 *   </svg>`
 * });
 *
 * // Legacy usage
 * const svgLegacy = Svg({ svg: "<svg>...</svg>" });
 * ```
 */
export declare function Svg(props: SvgProps): Element<ComponentInstance>;
export {};
//# sourceMappingURL=Graphic.d.ts.map