import { Geometry, Shader, Texture } from 'pixi.js';
import { Element } from '../engine/reactive';
import { ComponentInstance } from './DisplayObject';
import { DisplayObjectProps } from './types/DisplayObject';
import { SignalOrPrimitive } from './types';
import { ComponentFunction } from '../engine/signal';

/**
 * Interface defining the properties for a Mesh component.
 * Extends DisplayObjectProps to inherit common display object properties.
 */
interface MeshProps extends DisplayObjectProps {
    /** The geometry defining the mesh structure (vertices, indices, UVs, etc.) */
    geometry?: Geometry;
    /** The shader to render the mesh with */
    shader?: Shader;
    /** The texture to apply to the mesh */
    texture?: Texture | string;
    /** The image URL to load as texture */
    image?: string;
    /** The tint color to apply to the mesh */
    tint?: SignalOrPrimitive<number>;
    /** Whether to round pixels for sharper rendering */
    roundPixels?: SignalOrPrimitive<boolean>;
}
declare const CanvasMesh_base: {
    new (): {
        [x: string]: any;
        "__#1@#canvasContext": {
            [key: string]: any;
        } | null;
        isFlex: boolean;
        fullProps: import('..').Props;
        isMounted: boolean;
        _anchorPoints: import('pixi.js').ObservablePoint;
        isCustomAnchor: boolean;
        displayWidth: import('@signe/reactive').WritableSignal<number>;
        displayHeight: import('@signe/reactive').WritableSignal<number>;
        overrideProps: string[];
        layout: any;
        onBeforeDestroy: import('./DisplayObject').OnHook | null;
        onAfterMount: import('./DisplayObject').OnHook | null;
        subjectInit: import('rxjs').BehaviorSubject<any>;
        disableLayout: boolean;
        "__#1@#registeredEvents": Map<string, Function>;
        readonly deltaRatio: any;
        readonly parentIsFlex: any;
        onInit(props: import('..').Props): void;
        onMount({ parent, props }: Element</*elided*/ any>, index?: number): Promise<void>;
        onUpdate(props: import('..').Props): void;
        onDestroy(parent: Element, afterDestroy?: () => void): Promise<void>;
        setFlexDirection(direction: import('./types/DisplayObject').FlexDirection): void;
        setFlexWrap(wrap: "wrap" | "nowrap" | "wrap-reverse"): void;
        setAlignContent(align: import('./types/DisplayObject').AlignContent): void;
        setAlignSelf(align: import('./types/DisplayObject').AlignContent): void;
        setAlignItems(align: import('./types/DisplayObject').AlignContent): void;
        setJustifyContent(justifyContent: "flex-start" | "flex-end" | "center" | "space-between" | "space-around"): void;
        setPosition(position: import('./types/DisplayObject').EdgeSize): void;
        setX(x: number): void;
        setY(y: number): void;
        setPadding(padding: import('./types/DisplayObject').EdgeSize): void;
        setMargin(margin: import('./types/DisplayObject').EdgeSize): void;
        setGap(gap: import('./types/DisplayObject').EdgeSize): void;
        setBorder(border: import('./types/DisplayObject').EdgeSize): void;
        setPositionType(positionType: "relative" | "absolute"): void;
        setWidth(width: number): void;
        setHeight(height: number): void;
        getWidth(): number;
        getHeight(): number;
        setMinWidth(minWidth: number | string): void;
        setMinHeight(minHeight: number | string): void;
        setMaxWidth(maxWidth: number | string): void;
        setMaxHeight(maxHeight: number | string): void;
        setAspectRatio(aspectRatio: number): void;
        setFlexGrow(flexGrow: number): void;
        setFlexShrink(flexShrink: number): void;
        setFlexBasis(flexBasis: number | string): void;
        setRowGap(rowGap: number): void;
        setColumnGap(columnGap: number): void;
        setTop(top: number | string): void;
        setLeft(left: number | string): void;
        setRight(right: number | string): void;
        setBottom(bottom: number | string): void;
        setObjectFit(objectFit: import('./types/DisplayObject').ObjectFit): void;
        setObjectPosition(objectPosition: import('./types/DisplayObject').ObjectPosition): void;
        setTransformOrigin(transformOrigin: import('./types/DisplayObject').TransformOrigin): void;
    };
    [x: string]: any;
};
/**
 * Canvas Mesh component class that extends DisplayObject with PixiMesh functionality.
 * This component allows rendering of custom 3D meshes with shaders and textures.
 *
 * @example
 * ```typescript
 * // Basic mesh with geometry and texture
 * const mesh = Mesh({
 *   geometry: myGeometry,
 *   texture: "path/to/texture.png",
 *   tint: 0xff0000
 * });
 *
 * // Mesh with custom shader
 * const customMesh = Mesh({
 *   geometry: myGeometry,
 *   shader: myCustomShader,
 *   draw: (mesh) => {
 *     // Custom mesh manipulation
 *     mesh.rotation += 0.01;
 *   }
 * });
 * ```
 */
declare class CanvasMesh extends CanvasMesh_base {
    /**
     * Constructor for the CanvasMesh component.
     * Initializes the PixiMesh with default geometry and shader to prevent errors.
     *
     * @example
     * ```typescript
     * // This constructor is called internally by the engine
     * const mesh = new CanvasMesh();
     * ```
     */
    constructor();
    /**
     * Initializes the mesh component with the provided properties.
     * This method is called before onUpdate to set up initial state.
     *
     * @param props - The initial properties
     * @example
     * ```typescript
     * // This method is called internally when the component is created
     * mesh.onInit({
     *   geometry: myGeometry,
     *   texture: "texture.png"
     * });
     * ```
     */
    onInit(props: MeshProps): void;
    /**
     * Updates the mesh component when properties change.
     * Handles texture loading, shader updates, and other property changes.
     *
     * @param props - The updated properties
     * @example
     * ```typescript
     * // This method is called internally when props change
     * mesh.onUpdate({
     *   tint: 0x00ff00,
     *   texture: "new-texture.png"
     * });
     * ```
     */
    onUpdate(props: MeshProps): Promise<void>;
    /**
     * Called when the component is about to be destroyed.
     * Cleans up the draw effect subscription and calls the parent destroy method.
     *
     * @param parent - The parent element
     * @param afterDestroy - Callback function to execute after destruction
     * @example
     * ```typescript
     * // This method is typically called by the engine internally
     * await mesh.onDestroy(parentElement, () => console.log('Mesh destroyed'));
     * ```
     */
    onDestroy(parent: Element<ComponentInstance>, afterDestroy: () => void): Promise<void>;
}
/**
 * Creates a Mesh component with the specified properties.
 * This is the main function used to create mesh instances in your application.
 *
 * @param props - The properties for the mesh component
 * @returns A mesh component element
 * @example
 * ```typescript
 * import { Mesh } from 'canvasengine';
 *
 * // Create a basic textured mesh
 * const myMesh = Mesh({
 *   geometry: triangleGeometry,
 *   texture: "assets/texture.png",
 *   x: 100,
 *   y: 100,
 *   tint: 0xff0000
 * });
 *
 * // Create a mesh with custom shader
 * const shaderMesh = Mesh({
 *   geometry: planeGeometry,
 *   shader: customShader,
 *   draw: (mesh) => {
 *     mesh.rotation += 0.01;
 *   }
 * });
 * ```
 */
export declare const Mesh: ComponentFunction<MeshProps>;
export { CanvasMesh };
export type { MeshProps };
//# sourceMappingURL=Mesh.d.ts.map