import { WritableSignal } from '@signe/reactive';

export interface AnimateOptions<T> {
    duration?: number;
    ease?: (t: number) => number;
    onUpdate?: (value: T) => void;
    onComplete?: () => void;
}
export interface AnimatedState<T> {
    current: T;
    start: T;
    end: T;
}
export interface AnimatedSignal<T> extends Omit<WritableSignal<T>, 'set'> {
    (): T;
    set: (newValue: T, options?: AnimateOptions<T>) => Promise<void>;
    animatedState: WritableSignal<AnimatedState<T>>;
    update: (updater: (value: T) => T) => void;
}
export declare function isAnimatedSignal(signal: WritableSignal<any>): boolean;
/**
 * Creates an animated signal with the given initial value and animation options.
 * It's a writable signal that can be animated using popmotion. Properties of the animated signal are:
 * - current: the current value of the signal.
 * - start: the start value of the animation.
 * - end: the end value of the animation.
 *
 * @param initialValue The initial value of the signal.
 * @param options The animation options.
 * @returns The animated signal.
 * @example
 * const animatedValue = animatedSignal(0, { duration: 1000 });
 * animatedValue.set(10);
 * animatedValue.update((value) => value + 1);
 * console.log(animatedValue()); // 11
 *
 * animatedValue.animatedState() // { current: 10, start: 10, end: 11 }
 */
export declare function animatedSignal<T>(initialValue: T, options?: AnimateOptions<T>): AnimatedSignal<T>;
/**
 * Executes a sequence of animations. If an array is provided as an element in the sequence,
 * those animations will be executed in parallel.
 *
 * @param sequence Array of animation functions or arrays of animation functions for parallel execution
 * @returns Promise that resolves when all animations are complete
 * @example
 * ```ts
 * await animatedSequence([
 *   () => value1.set(10),
 *   [
 *     () => value2.set(20),
 *     () => value3.set(30)
 *   ],
 *   () => value1.set(0)
 * ])
 * ```
 */
export declare function animatedSequence(sequence: ((() => Promise<void>) | (() => Promise<void>)[])[]): Promise<void>;
//# sourceMappingURL=animation.d.ts.map