import { Signal } from '@signe/reactive';
import { Observable, Subject, Subscription } from 'rxjs';
import { ComponentInstance } from '../components/DisplayObject';
import { Directive } from './directive';

export interface Props {
    [key: string]: any;
}
type NestedSignalObjects = {
    [Key in string]: NestedSignalObjects | Signal<any>;
};
export interface Element<T = ComponentInstance> {
    tag: string;
    props: Props;
    componentInstance: T;
    propSubscriptions: Subscription[];
    effectSubscriptions: Subscription[];
    effectMounts: (() => void)[];
    effectUnmounts: ((element?: Element) => void)[];
    propObservables: NestedSignalObjects | undefined;
    parent: Element | null;
    context?: {
        [key: string]: any;
    };
    directives: {
        [key: string]: Directive;
    };
    destroy: () => void;
    allElements: Subject<void>;
}
type FlowResult = {
    elements: Element[];
    prev?: Element;
    fullElements?: Element[];
};
type FlowObservable = Observable<FlowResult>;
export declare const isElement: (value: any) => value is Element;
export declare const isPrimitive: (value: any) => boolean;
export declare function registerComponent(name: any, component: any): void;
/**
 * Creates a virtual element or a representation thereof, with properties that can be dynamically updated based on BehaviorSubjects.
 *
 * @param {string} tag - The tag name of the element to create.
 * @param {Object} props - An object containing properties for the element. Each property can either be a direct value
 *                         or an array where the first element is a function that returns a value based on input parameters,
 *                         and the second element is an array of BehaviorSubjects. The property is updated dynamically
 *                         using the combineLatest RxJS operator to wait for all BehaviorSubjects to emit.
 * @returns {Object} An object representing the created element, including tag name and dynamic properties.
 */
export declare function createComponent(tag: string, props?: Props): Element;
/**
 * Observes a BehaviorSubject containing an array or object of items and dynamically creates child elements for each item.
 *
 * @param {WritableArraySignal<T> | WritableObjectSignal<T>} itemsSubject - A signal that emits an array or object of items.
 * @param {Function} createElementFn - A function that takes an item and returns an element representation.
 * @returns {Observable} An observable that emits the list of created child elements.
 */
export declare function loop<T>(itemsSubject: any, createElementFn: (item: T, index: number | string) => Element | null): FlowObservable;
/**
 * Conditionally creates and destroys elements based on condition signals with support for else if and else.
 *
 * @description This function creates conditional rendering with support for multiple conditions (if/else if/else pattern).
 * It evaluates conditions in order and renders the first matching condition's element.
 * The function maintains full reactivity with signals and ensures proper cleanup of elements.
 *
 * @param {Signal<boolean> | boolean | (() => boolean)} condition - A signal, boolean, or function that determines whether to create an element.
 * @param {Function} createElementFn - A function that returns an element or a promise that resolves to an element.
 * @param {...Array} additionalConditions - Additional conditions for else if and else cases.
 *   Can be:
 *   - A function for else case: `() => Element | Promise<Element>`
 *   - An array for else if case: `[Signal<boolean> | boolean | (() => boolean), () => Element | Promise<Element>]`
 * @returns {Observable} An observable that emits the created element based on the matching condition.
 *
 * @example
 * ```typescript
 * // Simple if/else
 * cond(
 *   signal(isVisible),
 *   () => h(Container),
 *   () => h(Text, { text: 'Hidden' }) // else
 * );
 *
 * // Multiple else if + else
 * cond(
 *   signal(status === 'loading'),
 *   () => h(LoadingSpinner),
 *   [signal(status === 'error'), () => h(ErrorMessage)], // else if
 *   [signal(status === 'success'), () => h(SuccessMessage)], // else if
 *   () => h(DefaultMessage) // else
 * );
 * ```
 */
export declare function cond(condition: Signal<boolean> | boolean | (() => boolean), createElementFn: () => Element | Promise<Element>, ...additionalConditions: Array<(() => Element | Promise<Element>) | [Signal<boolean> | boolean | (() => boolean), () => Element | Promise<Element>]>): FlowObservable;
export {};
//# sourceMappingURL=reactive.d.ts.map