import { Subscription } from 'rxjs';
import { Element } from './reactive';
import { Tick } from '../directives/Scheduler';

type MountFunction = (fn: (element: Element) => void) => void;
export type ComponentFunction<P = {}> = (props: P) => Element | Promise<Element>;
export declare let currentSubscriptionsTracker: ((subscription: Subscription) => void) | null;
export declare let mountTracker: MountFunction | null;
/**
 * Registers a mount function to be called when the component is mounted.
 * To unmount the component, the function must return a function that will be called by the engine.
 *
 * @param {(element: Element) => void} fn - The function to be called on mount.
 * @example
 * ```ts
 * mount((el) => {
 *   console.log('mounted', el);
 * });
 * ```
 * Unmount the component by returning a function:
 * ```ts
 * mount((el) => {
 *   console.log('mounted', el);
 *   return () => {
 *     console.log('unmounted', el);
 *   }
 * });
 * ```
 */
export declare function mount(fn: (element: Element) => void): void;
/**
 * Registers a tick function to be called on each tick of the component's context.
 * @param {(tickValue: Tick, element: Element) => void} fn - The function to be called on each tick.
 * @example
 * ```ts
 * tick((tickValue, el) => {
 *   console.log('tick', tickValue, el);
 * });
 * ```
 */
export declare function tick(fn: (tickValue: Tick, element: Element) => void): void;
/**
 * Add tracking for subscriptions and mounts, then create an element from a component function.
 * @template C
 * @param {C} componentFunction - The component function to create an element from.
 * @param {Parameters<C>[0]} [props={}] - The props to pass to the component function.
 * @param {...any[]} children - The children elements of the component.
 * @returns {ReturnType<C>}
 * @example
 * ```ts
 * const el = h(MyComponent, {
 *   x: 100,
 *   y: 100,
 * });
 * ```
 *
 * with children:
 * ```ts
 * const el = h(MyComponent, {
 *   x: 100,
 *   y: 100,
 * },
 *   h(MyChildComponent, {
 *     x: 50,
 *     y: 50,
 *   }),
 * );
 * ```
 */
export declare function h<C extends ComponentFunction<any>>(componentFunction: C | Element, props?: Parameters<C>[0], ...children: any[]): ReturnType<C>;
export {};
//# sourceMappingURL=signal.d.ts.map