/**
 * Converts props into reactive signals if they are primitive values.
 *
 * @param {object} props - The properties to convert.
 * @param {object} [defaults={}] - Default values for properties.
 * @returns {object} An object with reactive signals.
 *
 * @example
 * const props = useProps({ count: 0, name: "John" });
 * console.log(props.count()); // 0
 * props.count.set(1);
 * console.log(props.count()); // 1
 */
export declare const useProps: (props: any, defaults?: {}) => any;
type PropType = NumberConstructor | StringConstructor | BooleanConstructor | FunctionConstructor | ObjectConstructor | ArrayConstructor | null | (new (...args: any[]) => any);
interface PropConfig {
    type?: PropType | PropType[];
    required?: boolean;
    default?: any | ((props: any) => any);
    validator?: (value: any, props: any) => boolean;
}
type PropSchema = {
    [key: string]: PropType | PropType[] | PropConfig;
};
/**
 * Validates and defines properties based on a schema.
 *
 * @param {object} props - The properties to validate.
 * @returns {function} A function that takes a schema and returns validated properties.
 *
 * @example
 * const schema = {
 *   age: { type: Number, required: true },
 *   name: { type: String, default: "Anonymous" }
 * };
 * const validatedProps = useDefineProps({ age: 25 })(schema);
 * console.log(validatedProps.age()); // 25
 * console.log(validatedProps.name()); // "Anonymous"
 */
export declare const useDefineProps: (props: any) => (schema?: PropSchema) => any;
export {};
//# sourceMappingURL=useProps.d.ts.map